/* delay - counts down a specified number of seconds.
   Copyright (C) 1998 Tom Rothamel.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  
*/

#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>

/* The following code implements the actual delaying. If precise is true, it
   attempts to compensate for the time spend in calculating, printing, etc.
*/

struct timeval tv;

void initdelay(int precise) {
	if (!precise) return;
	gettimeofday(&tv, NULL);
}

void delay(int secs, int precise) {
	struct timeval ctv;
	struct timeval dtv;

	if (!precise) {
		sleep(secs);
		return;
	}

	tv.tv_sec += secs;

	gettimeofday(&ctv, NULL);
	dtv.tv_sec  = tv.tv_sec - ctv.tv_sec;
	dtv.tv_usec  = tv.tv_usec - ctv.tv_usec;

	if (dtv.tv_usec < 0) {
		dtv.tv_usec += 1000000;
		dtv.tv_sec -= 1;
	}

	select(0, NULL, NULL, NULL, &dtv);
}

/* This converts a time offset to a number of seconds. For example, one
   can type 5:00 for a five minute delay, rather than typing 300 (the number
   of seconds.)

   It could be a little smarter.
*/

int atoseconds(char *a) {
	int i;
	int m = 1;
	int c = 0;
	int t = 0;
	
	for (i = strlen(a); i >= 0; i--) {
		switch(a[i]) {
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
		case '0':
			t += (a[i] - '0') * m;
			m *= 10;
			break;
		case ':':
			switch(c++) {
			case 0:
				m = 60;
			        break;
			case 1:
				m = 3600;
				break;
			case 2:
				m = 86400;
				break;
			}
		}			
	}

	return t;
}

/* Prints out the time, using the format code. The format is somewhat like
   the integer codes used in printf. You have all the flags, and then
   d for days, h for hours, m for minutes, s for seconds, and n for the
   total number of seconds remaining.
*/

void printtime(char *fmt, int time) {
	char buf[16];
	char *c;
	int i;
	char b;
	
	c = fmt;
	
	
	while (*c) {
		switch (*c) {
		case '%':
			strncpy(buf, c, 15);
			for (i = 1; i < 100; i++) {
				 if (buf[i] == 0) break;
				 if (buf[i] == 'd') break;
				 if (buf[i] == 'h') break;
				 if (buf[i] == 'm') break;
				 if (buf[i] == 's') break;
				 if (buf[i] == 'n') break;
			}

			b = buf[i];
			buf[i] = 'd';
			buf[i+1] = 0;

			switch (b) {
			case 'd':
				printf(buf, time/86400);
				break;
			case 'h':
				printf(buf, (time/3600) % 24);
				break;
			case 'm':
				printf(buf, (time/60) % 60);
				break;
			case 's':
				printf(buf, time % 60);
				break;
			case 'n':
				printf(buf, time);
				break;
			}
			
			c += i;
			break;
		default:
			printf("%c", *c);
			break;
		}

		c++;
	}

}


/* The rest of the code is responsible for displaying the countdown,
   parsing the arguments, and otherwise making this thing work.
*/

char *custmessage;

void usage() {
	fprintf(stderr, "Usage: delay [-hv] [-c custom]"
		        " <seconds to delay>\n");
	exit(-1);
}

void showcount(int dtime, int ctype) {
	switch(ctype) {
	case 0:
		printf("\r% 8d", dtime);
		break;
	case 1:
		printtime("\rTime Remaining: %d days, %02h:%02m:%02s.", dtime);
		break;
	case 2:
		printf("\r");
		printtime(custmessage, dtime);
	}

	fflush(stdout);
}

int main(int argc, char **argv) {
	char opt;
	int dtime;
	int ctype = 0;
	
	while (1) {
		opt = getopt(argc, argv, "hvpc:V"); 
		if (opt == EOF) break;
	
		switch (opt) {
		case 'h':
			usage();
			break;
		case 'v':
		        ctype = 1;
			break;
		case 'c':
			custmessage = optarg;
			ctype=2;
			break;
		case 'V':
			printf("delay - Version " VERSION "\n"
			       "  Copyright (c) 1998 Tom Rothamel\n"
			       "  This program has ABSOLUTELY NO WARRANTY. It can be distributed under\n"
			       "  the terms of the GNU General Public License.\n"); 
			exit(0);
		}

	}

	if (optind >= argc) {
		fprintf(stderr, "Error: You must supply a time to delay.\n");
		usage();
	}

	dtime = atoseconds(argv[optind]);
	
	if (dtime < 0) {
		fprintf(stderr, "Error: You've specified a negative delay time.\n");
		usage();
	}
	
	initdelay(1);
	
	while (dtime) {
		showcount(dtime, ctype);
		
		delay(1, 1); /* One second, always precise. */
		dtime--;
	}

	showcount(dtime, ctype);
	printf("\n");
	
	exit(0);
}
